# psf: ta=4, -*-Python-*-, vi: set et ts=4: coding: utf-8
__doc__ = """
ps_shell_log writes strings to the ps logging of the DEV_STAGE.
It should be used from accompanying Programms e.g. shell-scripts to
write logging Messages which will be integrated to the other logging
messages on the correlated DEV_STAGE.

  Usage: ps_shell_log [-v] [ --level=LEVEL ] [ --service_name=name ] \
            [ --have_config_file=true/false]  TEXT ...

  Arguments:
    TEXT  Message to be printed

  Options:
    -h --help             Show this screen.
    -v                    verbose
    --level=LEVEL         LogLevel (DEBUG,INFO,ERROR,WARNING,FATAL)
    --service_name=value   the system_id of the message

"""
import sys

import docopt

from ps.basic import Config
from ps.basic import __version__ as version

SERVICE_NAME = "shell_log"
VERBOSE = False


def main():
    """Add Log to logging stream."""
    global SERVICE_NAME, VERBOSE
    options = docopt.docopt(__doc__, version=version)
    if options["-v"]:
        VERBOSE = True
    if options["--have_config_file"] is not None:
        assert options["--have_config_file"] in ["True", "False"]
        have_config_file = eval(options["--have_config_file"])
    else:
        have_config_file = False
    if options["--level"] is not None:
        level = options["--level"]
    else:
        level = "info"
    if options["--service_name"] is not None:
        SERVICE_NAME = options["--service_name"]

    try:
        the_singleton = Config.Basic(
            SERVICE_NAME, have_config_file=have_config_file
        )

        if VERBOSE:
            the_singleton.verbose()
        exec(
            'Config.logger.%s("%s",extra={"package_version":"%s"})'
            % (level.lower(), options["TEXT"], str(version))
        )
    except Config.ForbiddenInitialisationOfSingleton:
        s = "Exception while calling ps_shell_log"
        sys.stderr.write(s)
        Config.logger.exception(s)



"""pytest file to store fixtures."""
import os

from ps.basic import DEV_STAGES

import pytest


@pytest.fixture(scope="module", params=DEV_STAGES.keys())
def dev_allowed_stages(request):
    os.environ["DEV_STAGE"] = request.param
    return request.param


@pytest.fixture(scope="module", params=["IMPOSSIBLE", "None", "EMPTY"])
def dev_wrong_stages(request):
    print("JUHU" + request.param)
    if request.param == "IMPOSSIBLE":
        os.environ["DEV_STAGE"] = request.param
    elif request.param == "None":
        os.environ.pop("DEV_STAGE")
    elif request.param == "EMPTY":
        os.environ["DEV_STAGE"] = ""
    else:
        raise
    return request.param



"""[summary]

Provide helper functions used within the test environment.
"""
import logging
import os
from importlib import reload
from socket import gethostname

import ps
from ps.basic import Config, DEV_STAGES

TEST_SERVICE_NAME = "test_service_name"


def reset_singleton(remove_log_file=True):
    """Reset the singleton's state initial values while testing."""
    if remove_log_file:
        if os.path.isfile(ps.basic.Config.log_file_name):
            os.remove(ps.basic.Config.log_file_name)
    Config = reload(ps.basic.Config)  # noqa: N806
    Config.Basic._instance = None
    properties = [
        "service_name",
        "suffix",
        "logging_port",
        "logging_bridge_port",
        "logging_level",
        "log_file_name",
        "logger",
        "config_file_name",
        "config_parser",
        "dev_stage",
        "herald_sqlite_filename",
        "primary_herald_url",
        "lock_file_name",
        "curr_mail_sender",
        "curr_mail_recipients",
        "curr_mail_subject",
        "curr_mail_text",
    ]
    for property in properties:
        assert Config.__dict__["%s" % (property)] == "not_yet_defined"
    logging.shutdown()
    reload(logging)


def get_data_of_file(filename_p: str):
    with open(filename_p, "r+t") as fp:
        data = fp.read()
        lines = data.split("\n")
        for line in lines:
            print(line)
    return data


def common_Config_class_attributes_after_initialisation(  # noqa: N802
    stage_p: str, test_service_name: str = TEST_SERVICE_NAME
):
    assert Config.service_name == test_service_name
    assert Config.__dict__["service_name"] == test_service_name
    assert os.getenv("SYSTEM_ID", None) == test_service_name
    assert os.getenv("SUB_SYSTEM_ID", None) == gethostname()
    assert Config.dev_stage == stage_p
    assert Config.suffix == DEV_STAGES[stage_p]["suffix"]
    assert Config.log_file_name == os.path.join(
        "LOG", Config.service_name + Config.suffix + ".log"
    )
    assert Config.logging_port == DEV_STAGES[stage_p]["logging_port"]
    assert Config.webserver_port == DEV_STAGES[stage_p]["logging_port"] + 1
    assert (
        Config.logging_bridge_port
        == DEV_STAGES[stage_p]["logging_bridge_port"]  # noqa: W503
    )
    assert Config.logging_level == DEV_STAGES[stage_p]["logging_level"]
    assert Config.admin_mail == DEV_STAGES[stage_p]["l_admin_mail"]
    assert Config.herald_sqlite_filename == os.path.join(
        os.getcwd(), "herald.sqlite%s" % Config.suffix
    )
    assert Config.lock_file_name == os.path.join(
        os.getcwd(), "%s_lock%s" % (Config.service_name, Config.suffix)
    )

    assert os.path.isdir("LOG")
    assert os.path.isfile(Config.log_file_name)



"""Test Lock File handling."""
import os

from ps.basic import Config, DEV_STAGES

import pytest


from t_utils import (
    TEST_SERVICE_NAME,
    common_Config_class_attributes_after_initialisation,
    get_data_of_file,
    reset_singleton,
)

os.environ["IS_TESTING"] = "YES"

#
# LOCK FILE THINGS
#


def test_properties_are_set_using_guarded_by_lockfile(
    dev_allowed_stages,
):
    """[summary]

    Lock File does not exist at start.
    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    # if dev_allowed_stages == "DEVELOPMENT":
    #     print( f" Skip {dev_allowed_stages}")
    #     return
    # Make sure the lock file is deleted
    lock_filename = os.path.join(os.getcwd(),
                                 "%s_lock%s" %
                                 (TEST_SERVICE_NAME,
                                  DEV_STAGES[dev_allowed_stages]["suffix"]),
                                 )
    print(f"Lockfilename is {lock_filename}")
    if os.path.isfile(lock_filename):
        os.remove(lock_filename)
    reset_singleton()

    Config.Basic(TEST_SERVICE_NAME, guarded_by_lockfile=True)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert "started new lock guarded instance on" in get_data_of_file(
        Config.log_file_name
    )
    assert Config.i_have_lock
    assert os.path.isfile(Config.lock_file_name)
    assert str(os.getpid()) in get_data_of_file(Config.lock_file_name)
    # Reset the singleton (but leave the lockfile in the filesystem)
    # The instantiation should raise a
    # Config.LockedInitialisationOfSingleton Exception
    # because the pid is still the same
    assert os.path.isfile(lock_filename)
    reset_singleton()
    assert os.path.isfile(lock_filename)
    # time.sleep(1)
    with pytest.raises(Config.LockedInitialisationOfSingleton):
        Config.Basic(TEST_SERVICE_NAME, guarded_by_lockfile=True)
        print(get_data_of_file(Config.log_file_name))
    assert os.getpid() == int(get_data_of_file(Config.lock_file_name))

    # Write a wrong pid into the lock file ==> now the instantiation will
    #  detect, that the service
    # is NOT running:
    #    - remove the lock file
    #    - and  raise a Config.LockedInitialisationOfSingleton exception
    assert os.path.isfile(Config.lock_file_name)
    with open(Config.lock_file_name, "w") as fp:
        fp.write("999999")
        fp.flush()
    reset_singleton(remove_log_file=False)
    with pytest.raises(Config.LockedInitialisationOfSingleton):
        Config.Basic(TEST_SERVICE_NAME, guarded_by_lockfile=True)
    assert "Will Remove the lock fil" in get_data_of_file(Config.log_file_name)
    print("The lockFileName is X%sX" % (Config.lock_file_name))
    assert not os.path.isfile(Config.lock_file_name)
    assert not Config.i_have_lock  # ==> False

    # Now the singleton could be used as at the beginning of the test
    reset_singleton(remove_log_file=False)
    Config.Basic(TEST_SERVICE_NAME, guarded_by_lockfile=True)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    print("The 2lockFileName is X%sX" % (Config.lock_file_name))
    assert os.path.isfile(Config.lock_file_name)
    assert str(os.getpid()) in get_data_of_file(Config.lock_file_name)



"""Tests for the ps.basic.fsm."""
import os
import pprint

import pytest

# Import of the fsm modules has to be done after initialisation of the
# Config Module as that needs the logger (Config.logger) to be set up.
from ps.basic import Config  # noqa:  I100

os.environ["IS_TESTING"] = "YES"
os.environ["DEV_STAGE"] = "TESTING"
Config.Basic("TEST_SERVICE_NAME")

from ps.basic.State import State, StateError  # noqa: E402 I202
from ps.basic.fsm import FiniteStateMachine, TransitionError  # noqa: E402
from ps.basic.get_graph import get_graph  # noqa: E402


# from t_utils import TEST_SERVICE_NAME, reset_singleton


def f(state: State, context: dict) -> str:
    """[summary]

    test_no_initial_stage_raise_Attribute_error.
    :param state: [description]
    :type state: State
    :param context: [description]
    :type context: dict
    :return: [description]
    :rtype: str
    """
    context[state.name] = "visited"
    print(state.name)
    context[state.name] = state.name
    print(pprint.pformat(context))
    if state.name == "START":
        return "input"
    if state.name == "STAGE1":
        return "input2"
    if state.name == "FIN":
        return "ii"


def test_no_initial_stage_raise_attribute_error():
    """[summary]

    test_no_initial_stage_raise_Attribute_error.
    """
    stage1 = State("STAGE1", f)
    fsm = FiniteStateMachine("TheMachine")
    fsm.add_state([stage1])
    with pytest.raises(AttributeError):
        fsm.run({})


def test_input_not_in_alphabet_raises_transition_error():
    """[summary]

    test_no_initial_stage_raise_Attribute_error.
    """
    stage1 = State("STAGE1", f)
    fsm = FiniteStateMachine("TheMachine")
    fsm.add_state([stage1])
    with pytest.raises(TypeError):
        fsm.process(["a"])


def test_multiple_initial_stages_raise_transition_error():
    """[summary]

    test_no_initial_stage_raise_Attribute_error.
    """
    start = State(
        "START",
        f,
        initial=True,
    )
    stage1 = State(
        "STAGE1",
        f,
        initial=True,
    )
    fsm = FiniteStateMachine("TheMachine")
    with pytest.raises(TransitionError):
        fsm.add_state([start, stage1])


def test_attributes_are_set():
    """[summary]

    test_no_initial_stage_raise_Attribute_error.
    """
    #    #reset_singleton()
    #    #Config.Basic(TEST_SERVICE_NAME)
    final = State("FINAL", f, final=True)
    error = State(
        "ERROR",
        f,
        error=True,
        mail_addr='[{"A":"error_handler@mail.company"}]',
        default=final,
    )
    start = State("START", f, initial=True, default=error)
    stage1 = State("STAGE1", f, default=error)

    start["input"] = stage1
    stage1["input2"] = final
    error["go_back_to_start"] = start
    error["go_back_to_stage1"] = stage1
    kk = start["input"]
    print(kk)

    print(start)
    # start.compute_function(start,d)
    # stage1.compute_function(stage1,d)
    # error.compute_function(error,d)
    # final.compute_function(final,d)
    #

    fsm = FiniteStateMachine("TheMachine")
    fsm.add_state([start, stage1, error, final])
    fsm.run({})
    # assert False

    graph = get_graph(fsm)
    graph.draw("x.png", prog="dot")
    graph.draw("x.svg", prog="dot")


#
# Graph Printing
#
def test_get_graph_without_initi_state_raises_stateerror():
    """[summary]

    test_no_initial_stage_raise_Attribute_error.
    """
    final = State("FINAL", f, final=True)
    fsm = FiniteStateMachine("TheMachine")
    fsm.add_state([final])
    with pytest.raises(StateError):
        get_graph(fsm)



# psf: ta=4, -*-Python-*-, vi: set et ts=4: coding: utf-8
#
__doc__ = """
  This module holds patterns for the different:

   -  commands (Remote/local)
   -  Email messages (subject and text)
   -  logging messages

   It is here for backward compatibility only.

   Actual implementations should use config files to get those patterns.
"""

PATTERN_LANGUAGES = ["DE", "EN"]
SUBJECT_PATTERNS = {
    "DE": {
        "PROD_LOCKED": "%(snapshot_name)s LOCKED ",
        "NO_PID": "Fehler beim Transfer aus %(sourcedir_p)s von \
              %(snapshot_name)s",
        "NO_EMAIL_TO": "Fehler beim Transfer aus \
              %(sourcedir_p)s von %(snapshot_name)s",
        "SUCC_TRANSFER": "[SYSTEM-DATA] SUCCESS\
               %(user_identification)s:%(PID)s transfer ",
        "FAIL_TRANSFER": "[SYSTEM-DATA] FAILURE \
              %(user_identification)s:%(PID)s transfer ",
        "RSYNC_FAILED": "Der rsync Aufruf/transfer hatte einen Fehler. \
              Abort now.",
        "RSYNC_SUCC": "SUCCESS rsync uebermittelte \
              %(asia_cloud_host)s/%(srcdir_on_asia_host)\
              s ==>%(PATH_TO_LOCAL_MIRROR_DIR)s",
        "QUARANTINED_TRANSFER": "Fehler beim Viruscheck",
        "FAIL_PRODMON_NOTIFICATION": "Fehler beim prodmon Aufruf \
                %(theurl)s fuer %(dst)s ",
        "SUCC_PRODMON_NOTIFICATION": "Erfolgreicher prodmon Aufruf\
               %(theurl)s fuer %(dst)s ",
    },
    "EN": {
        "PROD_LOCKED": "%(snapshot_name)s LOCKED ",
        "NO_PID": "Error in Transfer from %(sourcedir_p)s \
            of %(snapshot_name)s",
        "NO_EMAIL_TO": "Error in Transfer from %(sourcedir_p)s \
            of %(snapshot_name)s",
        "SUCC_TRANSFER": "[SYSTEM-DATA] SUCCESS \
            %(user_identification)s:%(PID)s transfer ",
        "FAIL_TRANSFER": "[SYSTEM-DATA] FAILURE \
            %(user_identification)s:%(PID)s transfer ",
        "RSYNC_FAILED": "The rsync transfer \
            exited with an error. Abort now.",
        "RSYNC_SUCC": "SUCCESS rsync transfered \
            %(asia_cloud_host)s/%(srcdir_on_asia_host)s \
            ==>%(PATH_TO_LOCAL_MIRROR_DIR)s",
        "QUARANTINED_TRANSFER": "Suspicous transmission qurantined ",
        "FAIL_PRODMON_NOTIFICATION": "FAILURE %(dst)s for %(theurl)s",
        "SUCC_PRODMON_NOTIFICATION": "SUCCESS %(dst)s for %(theurl)s",
    },
}
TEXT_PATTERNS = {
    "DE": {
        "PROD_LOCKED": "%(snapshot_name)s LOCKED ",
        "NO_PID": "PID aus ~/body konnte nicht bestimmt werden<br> s",
        "NO_EMAIL_TO": "Empfaenger konnten nicht bestimmt werden. \
            Existiert  ~/mail?<br> ",
        "SUCC_TRANSFER": "%(PID)s<br>Kuerzel: %(Kuerzel)s<br>Bezeichnung:\
             %(Bezeichnung)s<br>%(MESSAGE)s<br>%(email_footer)s<br> <br>",
        "FAIL_TRANSFER": "%(PID)s<br>Kuerzel: %(Kuerzel)s<br>Bezeichnung: \
            %(Bezeichnung)s<br>%(MESSAGE)s<br>%(email_footer)s<br>  \
            ----  <br>%(retval)s <br> ",
        "RSYNC_FAILED": "Der rsync Aufruf/transfer hatte einen Fehler.\
             Abort now.",
        "RSYNC_SUCC": "SUCCESS rsync uebermittelte \
            %(asia_cloud_host)s/%(srcdir_on_asia_host)s \
            ==>%(PATH_TO_LOCAL_MIRROR_DIR)s",
        "QUARANTINED_TRANSFER": "Fehler beim Viruscheck <br> %(cmd)s",
        "FAIL_PRODMON_NOTIFICATION": "Fehler beim prodmon \
            Aufruf %(dst)s   fuer %(theurl)s %(html)s",
        "SUCC_PRODMON_NOTIFICATION": "Erfolgreicher prodmon \
            Aufruf %(dst)s fuer %(theurl)s %(html)s",
    },
    "EN": {
        "PROD_LOCKED": "%(snapshot_name)s LOCKED ",
        "NO_PID": "PID in ~/body not found <br> ",
        "NO_EMAIL_TO": "Unable to locate E-Mail Address.  \
            Does ~/mail exist?<br> ",
        "SUCC_TRANSFER": "%(PID)s<br>shortname: %(Kuerzel)s<br>title:  \
            %(Bezeichnung)s<br>%(MESSAGE)s<br>%(email_footer)s<br>",
        "FAIL_TRANSFER": "%(PID)s<br>shortname: %(Kuerzel)s<br>title:  \
            %(Bezeichnung)s<br>%(MESSAGE)s<br>%(email_footer)s<br> \
                     ----  <br>%(retval)s <br> ",
        "RSYNC_FAILED": "The rsync transfer exited with an error. \
            Abort now.",
        "RSYNC_SUCC": "SUCCESS rsync transfered \
            %(asia_cloud_host)s/%(srcdir_on_asia_host)s \
            ==>%(PATH_TO_LOCAL_MIRROR_DIR)s",
        "QUARANTINED_TRANSFER": "Suspicous transmission \
            qurantined  %(cmd)s ",
        "FAIL_PRODMON_NOTIFICATION": "FAILURE %(dst)s for \
            %(theurl)s %(html)s",
        "SUCC_PRODMON_NOTIFICATION": "SUCCESS %(dst)s for\
             %(theurl)s %(html)s",
    },
}

LOGGING_PATTERNS = {
    "DE": {
        "CREATE_INPROGRESS_DIR": "LOCK Verzeichnis %(name)s angelegt",
        "PROD_LOCKED": "%(snapshot_name)s LOCKED ",
        "NO_PID": "%(snapshot_name)s hatt keine PID:      in \
            %(sourcedir_p)s/%(snapshot_name)s/body",
        "NO_EMAIL_TO": "%(snapshot_name)s hat keine  \
            EMAIL_TO: in %(sourcedir_p)s/%(snapshot_name)s/mail",
        "SUCC_TRANSFER": "%(MESSAGE)s",
        "FAIL_TRANSFER": "[SYSTEM-DATA] \
            FAILURE %(user_identification)s:%(PID)s transfer ",
        "EMAIL_FOOTER": "Bei Fragen bitte an \
            _EP-Produktionsstrecke wenden.",
        "PREPARE_CALLED": "prepare aufgerufen, fand:    \
            %(l_product_snapshots)s )",
        "CREATED_CLOUD_LOCKFILE": "%(cloud_lockfile_name)s CREATED",
        "ERROR_RM_CLOUD_LOCKFILE": "FEHLER %(cloud_lockfile_name)s \
            kann nicht geloescht werden.",
        "ERROR_CREATED_CLOUD_LOCKFILE": "Fehler %(cloud_lockfile_name)s \
            konnte nicht angelegt werden",
        "ALREADY_IN_TRANSFER": "%(snapshot_name)s  bereits in \
            transfer - ???? ",
        "NOT_LOCKED": "%(snapshot_name)s  ist nicht gesperrt",
        "TRANSFER_SUCC": "SUCCESS: Transfer von \
            %(user_identification)s:%(sourcedir_p)s/%(snapshot_name)s \
                                     zur SYSTEM-EU Machine",
        "TRANSFER_FAIL": "FAILURE Transfer \
            von %(user_identification)s:%(sourcedir_p)s/%(snapshot_name)s \
                                     zur SYSTEM-EU Machine",
        "CLOUD_TRANSFER_LOCKED": "%(cloud_lockfile_name)s \
            existiert. Transfer gesperrt!",
        # TRANSFER
        "TRANSFER_STARTED": "Transfer von \
            %(path_to_product_dir)s initiiert",
        "TRANSFER_NO_PROD_FOUND": "%(path_to_product_dir)s \
            ist kein Verzeichnis - Abort",
        "INITIAL_TRANSFER": "Es ist ein initialer transfer  \
            des Produktes %(product_name)s",
        "NO_INITIAL_TRANSFER": "%(destination_dir)s/%(product_name)s-* \
            existiert  (%(dirname_found)s)",
        "ALREADY_TRANSFERED": "%(product_id_stamp_p)s bereits transferiert",
        "CHK_IF_TRANSFER_WAS_OK": "Pruefe, ob der vorherige  transfer \
            von %(product_id_stamp_p)s erfolgreich war.",
        "PRECEEDING_TRANSF_OK": "Der vorherige  transfer von \
            %(product_id_stamp_p)s war erfolgreich. Skip Retransmission",
        "PRECEEDING_TRANSF_FAIL": "FAILURE des vorherigen  transfers \
            von %(product_id_stamp_p)s ",
        "MAYBE_REMOVE_WORK": "Ggf. muss %(product_id_stamp_p)s auf \
            Zielrechner geloescht werden. ",
        "TRY_RETRANSMISSION": "Versuche Retransmission .... . ",
        "VERSION_IS_BASE_FOR_RSYNC": "Die neueste Version %(dirname_found)s,  \
            wird als rsync-basis fuer %(product_id_stamp_p)s genommen. ",
        "SSH_EXIT_CODE_VULNERABILITY": "cp -al %(dirname_found)s ==>  \
            %(product_id_stamp_p)s.",
        "RSYNC_SUCC": "SUCCESS rsync uebermittelte %(path_to_product_dir)s",
        "TOUCH_TRANSM_OK_FILE_SUCC": "SUCCESS anlegen OK Datei in \
            %(destination_dir)s/%(product_id_stamp_p)s",
        "TOUCH_TRANSM_OK_FILE_FAIL": "FAILURE.Anlegen remote \
            OK File. Abort Now.",
        "START_FREESPACE": "===== Bereinige Festplatte fuer \
            %(product_name)s falls moeglich ========",
        "NUMBER_OF_FOUND_OK_FILES": "Anzahl %(number_of_ok_files)d \
            gefundener OK-Files auf der remote site",
        "ERR_NUMBER_OF_FOUND_OK_FILES": "FAILURE beim zaehlen der \
            remote OK Dateien. Abort Now.",
        "AMOUNT_PRODUCT_DIRS_SUCC": "%(number_of_products_dirs)s product \
            Verzeichnisse auf der remote site gefunden.",
        "ERR_AMOUNT_PRODUCT_DIRS": "FAILURE. Zaehlen der \
            remote OK Dateien. Abort Now.",
        "RM_DATA_ON_REMOTE": "Loesche altes Verzeichnis \
            %(name_of_oldest_product_dir)s auf remote",
        "RM_DATA_ON_REMOTE_SUCC": "SUCCESS Loesche altes Verzeichnis \
            %(name_of_oldest_product_dir)s",
        "RM_DATA_ON_REMOTE_FAIL": "FAILURE Loesche altes Verzeichnis \
            %(name_of_oldest_product_dir)s",
        "CALL_HELP_FOR_CLEANUP": "Anzahl ok-Files ungleich der Anzahl \
            der Datenverzeichnisse - bitte aufraumen",
        "NOTHING_TO_CLEAN": "Es wurde kein Speicherplatz freigegeben.",
        "AVAILABLE_OK_FILES": "Die vorhanden ok Files: %(ok_files)s",
        "AVAILABLE_PRODUCT_DIRS": "Die vorhanden \
            Verzeichnisse %(l_product_dirs)s",
        "RSYNC_FAILED": "Der rsync Aufruf/transfer hatte \
            einen Fehler. Abort now.",
        "LANGUAGE_PATTERN_FAILED": "Fehler beim setzen der LANGUAGE_PATTERN. \
            Verwende default %(PATTERN_LANGUAGE)s.",
    },
    "EN": {
        "CREATE_INPROGRESS_DIR": "LOCK Directory %(name)s created",
        "PROD_LOCKED": "%(snapshot_name)s LOCKED ",
        "NO_PID": "%(snapshot_name)s has no PID:      \
            in file %(sourcedir_p)s/%(snapshot_name)s/body",
        "NO_EMAIL_TO": "%(snapshot_name)s has no EMAIL_TO: in \
            file %(sourcedir_p)s/%(snapshot_name)s/mail",
        "SUCC_TRANSFER": "%(MESSAGE)s",
        "FAIL_TRANSFER": "[SYSTEM-DATA] FAILURE \
            %(user_identification)s:%(PID)s transfer ",
        "EMAIL_FOOTER": "In case of questions ... \
            contact  xx@somewhere.com.",
        "PREPARE_CALLED": " prepare Called, found:    \
            %(l_product_snapshots)s )",
        "CREATED_CLOUD_LOCKFILE": "%(cloud_lockfile_name)s CREATED",
        "ERROR_RM_CLOUD_LOCKFILE": "FAILURE %(cloud_lockfile_name)s \
            could not be deleted.",
        "ERROR_CREATED_CLOUD_LOCKFILE": "FAILURE %(cloud_lockfile_name)s \
            NOT CREATED",
        "ALREADY_IN_TRANSFER": "%(snapshot_name)s  \
            already in transfer - ???? ",
        "NOT_LOCKED": "%(snapshot_name)s  is not locked",
        "TRANSFER_SUCC": "SUCCESS: Transfer from \
            %(user_identification)s:%(sourcedir_p)s/%(snapshot_name)s \
                                     to DEST Machine",
        "TRANSFER_FAIL": "FAILURE Transfer \
            from %(user_identification)s:%(sourcedir_p)s/%(snapshot_name)s \
                                     to DEST Machine",
        "CLOUD_TRANSFER_LOCKED": "%(cloud_lockfile_name)s \
            existits. Transfer blocked!",
        # TRANSFER
        "TRANSFER_STARTED": "Transfer of\
             %(path_to_product_dir)s Initiated",
        "TRANSFER_NO_PROD_FOUND": "%(path_to_product_dir)s Not a\
             file - Abort",
        "INITIAL_TRANSFER": "It is an initial transfer  \
            of product %(product_name)s",
        "NO_INITIAL_TRANSFER": "%(destination_dir)s/%(product_name)s-* \
            exists (%(dirname_found)s)",
        "ALREADY_TRANSFERED": "%(product_id_stamp_p)s already transfered",
        "CHK_IF_TRANSFER_WAS_OK": "Check, if the (before) transfer \
            of %(product_id_stamp_p)s was successfull",
        "PRECEEDING_TRANSF_OK": "The (before) transfer of \
            %(product_id_stamp_p)s was successfull. Skip Retransmission",
        "PRECEEDING_TRANSF_FAIL": "FAILURE of the (before) \
            transfer of %(product_id_stamp_p)s ",
        "MAYBE_REMOVE_WORK": "Maybe a remove of %(product_id_stamp_p)s \
            on destination host is necessary. ",
        "TRY_RETRANSMISSION": "Try Retransmission .... . ",
        "VERSION_IS_BASE_FOR_RSYNC": "The newest version \
            %(dirname_found)s,  will be the rsync-base \
            for %(product_id_stamp_p)s ",
        "SSH_EXIT_CODE_VULNERABILITY": "DO NOT CARE ON \
            SUCCESS cp -al %(dirname_found)s to  %(product_id_stamp_p)s ",
        "RSYNC_SUCC": "SUCCESS rsync transfering %(path_to_product_dir)s",
        "TOUCH_TRANSM_OK_FILE_SUCC": "SUCCESS creating OK file in \
            %(destination_dir)s/%(product_id_stamp_p)s",
        "TOUCH_TRANSM_OK_FILE_FAIL": "Creating Remote OK File Failed. \
            Abort Now.",
        "START_FREESPACE": "===== FREE DISK SPACE \
            for %(product_name)s IF POSSIBLE ========",
        "NUMBER_OF_FOUND_OK_FILES": "Found %(number_of_ok_files)d \
            OK-Files on the remote site",
        "ERR_NUMBER_OF_FOUND_OK_FILES": "Unable to count \
            the remote OK File. Abort Now.",
        "AMOUNT_PRODUCT_DIRS_SUCC": "Found %(number_of_products_dirs)s \
            product dirs on the remote site",
        "ERR_AMOUNT_PRODUCT_DIRS": "Unable to count the product dirs \
            on the remote site . Abort Now.",
        "RM_DATA_ON_REMOTE": "REMOVE DATA ON DESTINATION \
            %(name_of_oldest_product_dir)s",
        "RM_DATA_ON_REMOTE_SUCC": "SUCCESS REMOVE DATA ON DESTINATION \
            %(name_of_oldest_product_dir)s",
        "RM_DATA_ON_REMOTE_FAIL": "FAILURE REMOVE DATA ON \
            DESTINATION %(name_of_oldest_product_dir)s",
        "CALL_HELP_FOR_CLEANUP": "Amount of ok-Files does not \
            match amount of Data directories - please take care",
        "NOTHING_TO_CLEAN": "No disk space freed on remote machine.",
        "AVAILABLE_OK_FILES": "Available ok Files: %(ok_files)s",
        "AVAILABLE_PRODUCT_DIRS": "Available Data directories \
            %(l_product_dirs)s",
        "RSYNC_FAILED": "The rsync transfer exited with an \
            error. Abort now.",
        "LANGUAGE_PATTERN_FAILED": "Error setting PATTERN_LANGUAGE. \
            Use default %(PATTERN_LANGUAGE)s.",
    },
}

PATTERN_NAMES = [
    name for name in globals().keys() if name.endswith("PATTERNS")
]

LOCAL_SHELL_CMDS = {
    # USED IN PREPARE
    "L_TOUCH_CLOUD_LOCKFILE": "touch %(cloud_lockfile_name)s",
    "L_TOUCH_PROD_LOCKFILE": "touch %(product_lockfile_name)s",
    "L_ZIP_TAR_INTO_PRODUCT_DIR": "gzip %(sourcedir_p)s/ERROR/%(snapshot_name)s-\
         %(prepare_started_at)s.tar.gz \
         %(sourcedir_p)s/%(snapshot_name)s ",
    "L_CREATE_ZIP_TAR_IN_ERR_DIR": "cd %(sourcedir_p)s ; \
            tar cfvz %(sourcedir_p)s/ERROR/%(snapshot_name)s_\
                %(prepare_started_at)s.tar.gz  %(snapshot_name)s ;",
    "L_RM_OLD_TRANSMISSIONS": "rm -r  %(filename)s",
    "L_RM_PROD_TS_DIR_HU2HU": "rm -rf \
        %(path_to_import_dir)s/%(snapshot_name)s",
    "L_RM_PROD_TS_DIR": "rm -rf %(sourcedir_p)s/%(snapshot_name)s",
    "L_RM_PROD_X_DIR": "rm -rf %(path_to_import_dir)s/%(dir_name)s",
    "L_RM_PROD_LOCKFILE": "rm %(product_lockfile_name)s",
    "L_RM_CLOUD_LOCKFILE": "rm  %(cloud_lockfile_name)s",
    # USED IN TRANSFER
    "L_TOUCH_PROD_LOGFILE": "touch   %(logfile_name)s",
}

# Snapshotnames on the way to europe have a - (hyphen)  in
# between Pi-Name and timestamp.
# Snapshotnames on the way to china have a _
# (underscore) in between Pi-Name and timestamp.
# Therefore we distinguish _ and - versions for the
# following commands.
# The value (hyphen or underscore) is set via  the global
# variable DIRECTION_ISOLATOR.

REMOTE_SHELL_CMDS = {
    "R_FIND_YOUNGEST_PRODUCT_DIR": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s* \
        -maxdepth 0 -type d | tail -1",
    "R_FIND_YOUNGEST_PRODUCT_DIR": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s* \
        -maxdepth 0 -type d | tail -1",
    "R_FIND_OK_FILE_OF_CURRENT_TRANSMISSION": "find %(destination_dir)s/%(expected_name_of_ok_file)s \
        -maxdepth 0 -type f",
    "R_CP_OLD_TO_STATIC_LINKED_NEW": "cp -al %(dirname_found)s \
        %(destination_dir)s/%(product_id_stamp_p)s",
    "R_TOUCH_TRANSMISSION_OK_FILE": "touch %(destination_dir)s/%(product_id_stamp_p)s.ok",  # noqa: E501
    "R_GET_AMOUNT_PRODUCT_DIRS": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s* \
        -maxdepth 0 -type d | wc -l",
    "R_NAME_OF_OLDEST_PRODUCT_DIR": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s* \
        -maxdepth 0 -type d | head -1",
    "R_REMOVE_OLDEST_PRODUCT_DIR": "rm -rf %(name_of_oldest_product_dir)s*",
    "R_GET_AMOUNT_PRODUCT_OK_FILES": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s*ok \
        -maxdepth 0 -type f | wc -l",
    "R_GET_OLDEST_PRODUCT_DIR": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s* \
        -maxdepth 0 -type d | head -1",
    "R_RM_OLDEST_PRODUCT_DIR": "rm -rf %(oldest_data)s*",
    "R_LIST_PRODUCT_OK_FILES": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s*ok \
        -maxdepth 0 -type f",
    "R_LIST_PRODUCT_DIRS": "find %(destination_dir)s/%(product_name)s%(DIRECTION_ISOLATOR)s* \
        -maxdepth 0 -type d ",
    "R_FIND_CLEANME_DIRS": "find %(dstdir_on_eu_host)s -name PI*CLEANME  -print  ",  # noqa: E501
}



import inspect
from pprint import pformat

from ps.basic.State import StateError


def get_graph(fsm, title: str = None):
    """Generate a png/svg.

    Picture with dot - a graphviz executable.

    Have a look at the Usage section for a working example.

    :param fsm: [description]
    :type fsm: [type]
    :param title: [description], defaults to None
    :type title: str, optional
    :raises StateError: [description]
    :return: [description]
    :rtype: [type]
    """
    import pygraphviz as pgv

    fsm_dot_attrs = {
        "directed": True,
        "strict": False,
        "rankdir": "LR",
        "ratio": "0.3",
    }

    state_dot_attrs = {
        "shape": "circle",
        "height": "1.2",
    }

    dot_final = "doublecircle"

    title = fsm.name

    fsm_graph = pgv.AGraph(title=title, size="10.3, 5.3", **fsm_dot_attrs)
    fsm_graph.node_attr.update(state_dot_attrs)

    if fsm.init_state is None:
        raise StateError("A FSM must have a defined init_state")

    for state in [fsm.init_state] + fsm.states:
        color = "black"
        shape = state_dot_attrs["shape"]
        if hasattr(fsm, "final_states"):
            if id(state) in [id(s) for s in fsm.final_states]:
                shape = dot_final
        if hasattr(fsm, "error_states"):
            if id(state) in [id(s) for s in fsm.error_states]:
                color = "red"

        try:
            tooltip_str = pformat(inspect.getdoc(state.compute_function))
            state.tooltip = tooltip_str.replace("\\n", "&#013;")
        except AttributeError:  # pragma: no cover
            state.tooltip = "no tooltip"

        fsm_graph.add_node(
            n=state.name, shape=shape, color=color, tooltip=state.tooltip
        )
        if state.mail_addr:
            lis = eval(state.mail_addr)
            for d in lis:
                l2 = d.keys()
                name = list(l2)[0]
                value = d[name]
                fsm_graph.add_node(
                    n="EMAIL_DST" + name + value,
                    label="%s:%s" % (name, value),
                    shape="rpromoter",
                    style="filled",
                    fillcolor="lightseagreen:orangered",
                    color=color,
                    tooltip=state.tooltip,
                )
                fsm_graph.add_edge(
                    state.name,
                    "EMAIL_DST" + name + value,
                    penwidth=2,
                    label="send email ",
                    style="dashed",
                    color=color,
                    tooltip=state.tooltip,
                )

    fsm_graph.add_node("null", shape="plaintext", label=" ")
    fsm_graph.add_edge("null", fsm.init_state.name)

    for src, input_value, dst in fsm.all_transitions:
        label = str(input_value)
        fsm_graph.add_edge(src.name, dst.name, label=label)
    for state in fsm.states:
        if state.default_transition is not None:
            if id(state.default_transition) in \
                    [id(s) for s in fsm.error_states]:
                fsm_graph.add_edge(
                    state.name,
                    state.default_transition.name,
                    label="error",
                    color="red",
                    tooltip=state.tooltip,
                )
            else:
                fsm_graph.add_edge(
                    state.name,
                    state.default_transition.name,
                    label="else",
                    tooltip=state.tooltip,
                )

    return fsm_graph



"""Tests for the ps_shell_log."""
import sys

from ps.basic import ps_shell_log

from t_utils import (
    common_Config_class_attributes_after_initialisation,
    reset_singleton
)

SERVICE_NAME = "A_SERVICE_NAME"


def test_ps_1(dev_allowed_stages, capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    sys.argv = [
        "ps_shell_log",
        "-v",
        "--service_name",
        SERVICE_NAME,
        "--level",
        "info",
        "THE MESSAGE",
    ]
    ps_shell_log.main()
    common_Config_class_attributes_after_initialisation(
        dev_allowed_stages, test_service_name=SERVICE_NAME
    )
    captured = capsys.readouterr()
    assert "THE MESSAGE" in captured.err


def test_ps_shell_log_forced_configfile_not_given(dev_allowed_stages, capsys):
    """[summary]

    have_config_file is set to True BUT a config_file is not provided.
    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    sys.argv = [
        "ps_shell_log",
        "-v",
        "--service_name",
        SERVICE_NAME,
        "--have_config_file",
        "True",
        "THE MESSAGE",
    ]
    ps_shell_log.main()
    common_Config_class_attributes_after_initialisation(
        dev_allowed_stages, test_service_name=SERVICE_NAME
    )
    captured = capsys.readouterr()
    assert "Exception while calling ps_shell_log" in captured.err



"""Test State Module."""

import os
import pprint

from ps.basic.State import State, StateError

import pytest


os.environ["IS_TESTING"] = "YES"


def f(state: State, context: dict) -> str:
    """[summary]

    transition_func.
    :param state: [description]
    :type state: State
    :param context: [description]
    :type context: dict
    :return: [description]
    :rtype: str
    """
    context[state.name] = "visited"
    #    print(state.name)
    d = {}
    d[state.name] = state.name
    print(pprint.pformat(context))
    if state.name == "START":
        return "input"
    if state.name == "STAGE1":
        return "input2"
    if state.name == "FIN":
        return "ii"


def test_attributes_are_set():
    """[summary]

    test_attributes_are_set.
    """
    global f
    final = State("FINAL", f, final=True)
    assert final.name == "FINAL"
    assert final.compute_function == f
    assert not final.error  # ==> False
    assert not final.initial  # ==> False
    assert final.final  # ==> True
    assert final.default_transition is None


def test_default_transition_is_set():
    """[summary]

    test_default_transition_is_set.
    """
    global f
    final = State("FINAL", f, final=True)
    initial = State("INITIAL", f, initial=True, default=final)
    assert initial.default_transition == final


def test_strings_for_input_alphabet_could_be_set():
    """[summary]

    test_strings_for_input_alphabet_could_be_set.
    """
    global f
    final = State("FINAL", f, final=True)
    intermediate = State("INTERMEDIATE", f)
    intermediate["input2"] = final
    assert intermediate["input2"] == final


def test_wrong_result_state_on_input_raise_error():
    """[summary]

    test_wrong_result_state_on_input_raise_error.
    """
    with pytest.raises(StateError):
        intermediate = State("INTERMEDIATE", f)
        intermediate["input2"] = "xx"


def test_repr_works():
    """[summary]

    test_repr_works.
    """
    global f
    final = State("FINAL", f, final=True)
    x = repr(final)
    assert "FINAL" in x
    assert "State" in x



from typing import Any, Callable


class StateError(Exception):
    """[summary]

    Error in setting a destination state based on a string.

    :param Exception: [description]
    :type Exception: [type]
    """

    def __init__(self, message):
        """[summary]

        :param message: [description]
        :type message: [type]
        """
        self.message = message


class State(dict):
    """[summary]

        :raises StateError: [description]
        :return: [description]
        :rtype: [type]

    State class.

    A State instance is a dictionary:
         - its keys are the strings identified for state_changes of the
           fsm the states are used within
         - the value correlated to the key defines a state instance
           (the next state) if the state computing function returns that value

    We first define a state_enter_function *f* ...

    >>> def f(state: State, context_p: dict) -> str:
    ...     context_p["fin_visited"] = True
    ...     return "FIN"

    In the next step we define a State. The creation of the State
    indicates, that the function
    *f* will be executed if we enter the state.
    Furthermore we indicate, (final=True) that the state is a final state.

    >>> final = State("FINAL", f, final=True)

    We define a dictionary named context. That dictionary is a required
    parameter to any  states computing function.

    >>> context={}

    If we call the states computing function it will return the string "FIN"

    >>> final.compute_function(final,context)
    'FIN'

    It will modify the context

    >>> context["fin_visited"]
    True

    Let us create other states

    >>> other_state = State("other_state", f)
    >>> other_state2 = State("other_state2", f)

    and set it's behaviour so, that if its compute fuction returns
    'Go_To_Fi_State' it will change its state to the final state.

    >>> other_state["Go_To_Fi_State"] = final
    >>> other_state["Go_To_Other_State2"] = other_state2

    To show the state_change list we issue:

    >>> other_state.keys()
    dict_keys(['Go_To_Fi_State', 'Go_To_Other_State2'])

    To find out, to which state we change if the states compute function
    returns a value we check:

    >>> other_state['Go_To_Fi_State']  #doctest: +ELLIPSIS
    <'FINAL' State ...>
    >>> other_state['Go_To_Other_State2']  #doctest: +ELLIPSIS
    <'other_state2' State ...>

    """

    def __init__(
        self,
        name: str,
        compute_function: Callable[[Any, dict], str],
        initial: bool = False,
        final: bool = False,
        error: bool = False,
        mail_addr: str = "",
        default: Any = None,
    ):
        """[summary]

        Construct a state.
        :param name: [description]
        :type name: str
        :param compute_function: [description]
        :type compute_function: Callable[[Any, dict], str]
        :param initial: [description], defaults to False
        :type initial: bool, optional
        :param final: [description], defaults to False
        :type final: bool, optional
        :param error: [description], defaults to False
        :type error: bool, optional
        :param mail_addr: [description], defaults to ""
        :type mail_addr: str, optional
        :param default: [description], defaults to None
        :type default: Any, optional
        """
        self.name = name
        self.compute_function = compute_function
        self.final = final
        self.error = error
        self.initial = initial
        self.mail_addr = mail_addr
        self.default_transition = default
        self.tooltip = "None"

    def __getitem__(self, input_value):
        """[summary]

        Make a transition to the next state.
        :param input_value: [description]
        :type input_value: [type]
        :return: [description]
        :rtype: [type]
        """
        return dict.__getitem__(self, input_value)

    def __setitem__(self, input_value, next_state):
        """[summary]

        Set a transition to a new state.
        :param input_value: [description]
        :type input_value: [type]
        :param next_state: [description]
        :type next_state: [type]
        :raises StateError: [description]
        """
        if not isinstance(next_state, State):
            raise StateError(
                "A state must transition to another state,"
                " got %r instead." % next_state
            )
        dict.__setitem__(self, input_value, next_state)

    def __repr__(self):
        """[summary]

        Represent the object in a string.
        :return: [description]
        :rtype: [type]
        """
        return "<%r %s @ 0x%x>" % (
            self.name,
            self.__class__.__name__,
            id(self),
        )



"""Tests for the ps.basic.util Module."""
from ps.basic.utils import get_html_string, hms_string

import pytest


@pytest.mark.parametrize(
    "test_input1,expected1",
    [
        (
            {"A": "B"},
            "<table><tr><td>A</td><td>B</td></tr></table>",
        ),  # dictionary
        (
            {"A": {"B1": "B2"}},
            "<table><tr><td>A</td><td><table><tr><td>B1</td><td>B2</td></tr></table></td></tr></table>",  # noqa: E501
        ),
    ],
)
def test_html_string(test_input1, expected1):
    """[summary]

    :param test_input1: [description]
    :type test_input1: [type]
    :param expected1: [description]
    :type expected1: [type]
    """
    assert get_html_string(test_input1) == expected1


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (1, "00:00:01"),
        (61, "00:01:01"),
        (3661, "01:01:01"),
    ],
)
def test_hms_string(test_input, expected):
    """[summary]

    :param test_input: [description]
    :type test_input: [type]
    :param expected: [description]
    :type expected: [type]
    """
    assert hms_string(test_input) == expected



__doc__ = """ The Config Module delivers the *Basis* for Systems in the Production
     Systems world."""

import logging
import logging.handlers
import os
import signal
import sys
import tempfile
import time
import traceback
from configparser import (
    ConfigParser,
    NoOptionError,
    NoSectionError,
    ParsingError,
)
from socket import gethostname
from subprocess import PIPE, Popen

from ps.basic import DEV_STAGES, __version__, hms_string
from ps.basic.Patterns import LOGGING_PATTERNS, PATTERN_LANGUAGES

not_yet_defined = "not_yet_defined"


service_name = not_yet_defined
suffix = not_yet_defined
logging_port = not_yet_defined
logging_bridge_port = not_yet_defined
webserver_port = not_yet_defined
logging_level = not_yet_defined
log_file_name = not_yet_defined
logger = not_yet_defined
config_file_name = not_yet_defined
config_parser = not_yet_defined
dev_stage = not_yet_defined
herald_sqlite_filename = not_yet_defined
primary_herald_url = not_yet_defined
lock_file_name = not_yet_defined
admin_mail = not_yet_defined
curr_mail_sender = not_yet_defined
curr_mail_recipients = not_yet_defined
curr_mail_subject = not_yet_defined
curr_mail_text = not_yet_defined
is_testing = not_yet_defined
i_have_lock = False


MAX_SIZE_FOR_A_ROTATING_LOGFILE = 2000000
MAX_NUMBER_OF_OLD_ROTATING_LOGFILES = 10
PATTERN_LANGUAGE = "EN"


def sighup_handler(signum: int, frame):
    """[summary]

    :param signum: [description]
    :type signum: int
    :param frame: [description]
    :type frame: [type]
    :raises ForbiddenInitialisationOfSingleton: [description]
    """
    global config_parser, logger
    if config_parser != not_yet_defined:
        logger.debug(
            "SIGHUP Received. Print Configuration.",
            extra={"package_version": __version__},
        )
        log_config_data(config_parser, logger)
    else:
        raise ForbiddenInitialisationOfSingleton(
            "SIGHUP called but module not yet initiated"
        )


# Enable the signal handler
signal.signal(signal.SIGHUP, sighup_handler)


def log_config_data(config_parser: ConfigParser, logger: logging.Logger):
    """[summary]

    :param config_parser: [description]
    :type config_parser: ConfigParser
    :param logger: [description]
    :type logger: logging.Logger
    """
    html_string = "<table>"
    for section_name in config_parser.sections():
        html_string += "<tr><td>%s</td></tr>" % (section_name)
        for name, value in config_parser.items(section_name):
            html_string += "<tr><td>%s</td><td>%s</td></tr>" % (
                str(name),
                str(value),
            )
            #  name is lower() of the real name as default of config_parser
            if section_name == "GLOBAL" and name == "logging":
                numeric_level = getattr(logging, value.upper(), None)
                #                print(f"CONF2 Set Logging Level to {value}")
                logger.setLevel(numeric_level)
        html_string += "</table>"
        if html_string != "<table></table>":
            logger.debug(
                f"{html_string}", extra={"package_version": __version__}
            )


class Basic(object):
    """[summary]

    :param service_name_p: name of the service
    :type service_name_p: str
    :param have_config_file: [force usage of config file], defaults to False
    :type have_config_file: bool, optional
    :param guarded_by_lockfile: [create/check lockfile], defaults to False
    :type guarded_by_lockfile: bool, optional
    :param have_herald_url_in_config_file: [ force usage of \
        herald_url in config], defaults to False
    :type have_herald_url_in_config_file: bool, optional
    :raises ForbiddenInitialisationOfSingleton: [description]
    :raises ForbiddenInitialisationOfSingleton: [description]

    The Basic Class delivers the *Basis* for Systems in the Production System
    world.

       It enforces the same patterns for:
         - logging
         - configuration files
         - and maybe more

       throughout the different development stages.

       **Example Usage**

       >>> from ps.basic import Config
       >>> import os
       >>> project_name='ps'
       >>> dev_stage='DEVELOPMENT'
       >>> # DEV_STAGE has to be defined in the environment. Error else
       >>> os.environ["DEV_STAGE"]=dev_stage
       >>> # configuration files are built from the projects name
       >>> # and a shortcut for the development stage.
       >>> # They follow the ini syle (ms-world) for
       >>> # configuration files.
       >>> fp=open(os.path.join("/tmp", project_name + "_d.cfg"),"w")
       >>> written = fp.write("[GLOBAL]" + os.linesep)
       >>> written = fp.write("pattern_language=EN")
       >>> fp.close()
       >>> # you can set the directory, where config files live
       >>> os.environ["BASIC_CONFIGFILE_DIR"] = "/tmp"
       >>> singleton=Config.Basic(project_name,have_config_file = True)
       >>> Config.config_file_name
       '/tmp/ps_d.cfg'
       >>> assert(Config.config_file_name == "/tmp/ps_d.cfg")
       >>> my_version="0.0.alpha"
       >>> # Will leave an Info level log message in the local log file
       >>> # as in the (DEV_STAGE) central Log-Server/herald"
       >>> logger = Config.logger
       >>> logger.info("Hello LoggingWorld",
       ...      extra={"package_version":my_version})
       >>> # The application level access should occure via the
       >>> # provided class variables to read the config file
       >>> Config.config_parser.sections()
       ['GLOBAL']
       >>> Config.config_parser.items('GLOBAL')
       [('pattern_language', 'EN')]
       >>> # put logging messages to stdout too
       >>> singleton.verbose()

    """

    _instance = None

    def __init__(
        self,
        service_name_p: str,
        have_herald_url_in_config_file: bool = False,
        have_config_file: bool = False,
        guarded_by_lockfile: bool = False,
    ):
        """[summary]

        Create the singleton.
        """
        global service_name, dev_stage, suffix, logging_port
        global logging_bridge_port, webserver_port, logging_bridge_port
        global logging_level, log_file_name, logger, config_file_name
        global config_parser, herald_sqlite_filename, primary_herald_url
        global lock_file_name, admin_mail, curr_mail_sender
        global curr_mail_recipients, curr_mail_subject, curr_mail_text
        global is_testing, i_have_lock

        if Basic._instance is None:
            dev_stage = os.getenv("DEV_STAGE", None)
            if dev_stage is None or dev_stage not in DEV_STAGES.keys():
                raise ForbiddenInitialisationOfSingleton(
                    f"DEV_STAGE {dev_stage}  not an Allowed DEVELOPMENT STAGE"
                )
            Basic._instance = self
            service_name = service_name_p
            os.environ["SYSTEM_ID"] = service_name_p
            os.environ["SUB_SYSTEM_ID"] = gethostname()
            suffix = DEV_STAGES[dev_stage]["suffix"]
            log_file_name = os.path.join("LOG", service_name + suffix + ".log")
            logging_port = DEV_STAGES[dev_stage]["logging_port"]
            webserver_port = DEV_STAGES[dev_stage]["logging_port"] + 1
            logging_bridge_port = DEV_STAGES[dev_stage]["logging_bridge_port"]
            logging_level = DEV_STAGES[dev_stage]["logging_level"]
            admin_mail = DEV_STAGES[dev_stage]["l_admin_mail"]
            herald_sqlite_filename = os.path.join(
                os.getcwd(), f"herald.sqlite{suffix}"
            )
            primary_herald_url = ""
            lock_file_name = os.path.join(
                os.getcwd(), f"{service_name_p}_lock{suffix}"
            )
            is_testing = os.getenv("IS_TESTING", "NOT TESTING")

            # Create Local LOG Directory
            if not os.path.isdir("LOG"):
                try:
                    os.mkdir("LOG")
                except FileNotFoundError:  # pragma: no cover
                    sys.stderr.write(
                        f"Error creating LOG DIR in {os.getcwd()}: EXIT Now"
                    )
                    sys.stderr.write("Error creating LOG DIR: EXIT Now")
                    sys.exit(1)

            logger = logging.getLogger(service_name)
            my_filter = ContextFilter()
            logger.addFilter(my_filter)
            socket_handler = logging.handlers.SocketHandler(
                "localhost", logging_port
            )
            logger.addHandler(socket_handler)
            logfile_handler = logging.handlers.RotatingFileHandler(
                log_file_name,
                maxBytes=MAX_SIZE_FOR_A_ROTATING_LOGFILE,
                backupCount=MAX_NUMBER_OF_OLD_ROTATING_LOGFILES,
            )
            formatter = logging.Formatter(
                "%(asctime)s %(PRODUKT_ID)s %(SYSTEM_ID)s %(SUB_SYSTEM_ID)s\
                 %(SUB_SUB_SYSTEM_ID)s %(USER_SPEC_1)s %(USER_SPEC_2)s \
                 %(levelname)s %(name)-12s %(package_version)s \
                 %(lineno)d %(message)s"
            )
            logfile_handler.setFormatter(formatter)
            logger.addHandler(logfile_handler)

            logger.setLevel(logging_level)
            logger.debug(
                f"Logging initialized for pid {os.getpid()}",
                extra={"package_version": __version__},
            )

            if guarded_by_lockfile:
                self.__handle_lockfile__()

            if have_config_file:
                self.__handle_configfile__(have_herald_url_in_config_file)
        else:
            raise ForbiddenInitialisationOfSingleton(
                "Reinitialization of the singleton is not allowed"
            )

    def __handle_configfile__(self, have_herald_url_in_config_file: bool):
        """[summary]

        :param have_herald_url_in_config_file: [description]
        :type have_herald_url_in_config_file: bool
        :raises ForbiddenInitialisationOfSingleton: [description]
        :raises ForbiddenInitialisationOfSingleton: [description]
        :raises e: [description]
        :raises e: [description]
        :raises e: [description]
        :raises ForbiddenInitialisationOfSingleton: [description]
        :raises ForbiddenInitialisationOfSingleton: [description]
        """
        global config_parser, config_file_name, PATTERN_LANGUAGE
        global config_file_directory, primary_herald_url
        config_file_directory = os.getenv("BASIC_CONFIGFILE_DIR", False)
        if config_file_directory:
            if not os.path.isdir(config_file_directory):
                s = f"BASIC_CONFIGFILE_DIR {config_file_directory} not found."
                logger.fatal(
                    s,
                    extra={"package_version": __version__},
                )
                sys.stderr.write(s)
                raise ForbiddenInitialisationOfSingleton(s)
            else:
                config_file_name = os.path.join(
                    config_file_directory, service_name + suffix + ".cfg"
                )
                logger.info(
                    f"The config_file was forced  to {config_file_name}. ",
                    extra={"package_version": __version__},
                )
        else:
            config_file_name = os.path.join(
                os.getcwd(), service_name + suffix + ".cfg"
            )

        config_parser = ConfigParser()
        try:
            if os.path.isfile(config_file_name):
                config_parser.read(config_file_name)
            else:
                s = f"No config file {config_file_name} "
                s += "given, but it's usage is mandatory."
                logger.fatal(s, extra={"package_version": __version__})
                raise ForbiddenInitialisationOfSingleton(s)

        except (NoSectionError, ParsingError, NoOptionError) as e:
            e = sys.exc_info()[1]
            logger.exception(
                "Error reading config File: version %s" % (__version__)
            )
            sys.stderr.write("Error reading config File: EXIT Now")
            sys.stderr.write(traceback.format_exc())
            self.__exit__(1, 2, 3)
            if is_testing == "YES":
                raise e
            else:  # pragma: no cover
                raise e
                sys.exit(1)
        except SystemExit:  # pragma: no cover
            logger.exception(
                f"Error reading config File: version {__version__}"
            )
            e = sys.exc_info()[1]
            self.__exit__(1, 2, 3)
            raise e

        if have_herald_url_in_config_file:
            try:
                primary_herald_url = config_parser.get("GLOBAL", "herald_url")
            except (NoSectionError, ParsingError, NoOptionError):
                s = "Value of herald_url in GLOBAL section of "
                s += f"{config_file_name} not given - but needed. EXIT NOW"
                logger.fatal(
                    s,
                    extra={"package_version": __version__},
                )
                sys.stderr.write(s)
                raise ForbiddenInitialisationOfSingleton(s)
        try:
            PATTERN_LANGUAGE = config_parser.get("GLOBAL", "pattern_language")
            if PATTERN_LANGUAGE not in PATTERN_LANGUAGES:
                s = "Value of pattern_language in GLOBAL section of "
                s += f"{config_file_name}  not allowed. EXIT NOW"
                logger.fatal(
                    s,
                    extra={"package_version": __version__},
                )
                sys.stderr.write(s)
                raise ForbiddenInitialisationOfSingleton(s)
        except (NoSectionError, ParsingError, NoOptionError):
            PATTERN_LANGUAGE = "EN"
            d = {}
            d["PATTERN_LANGUAGE"] = PATTERN_LANGUAGE
            logger.debug(
                template_writer(
                    LOGGING_PATTERNS,
                    PATTERN_LANGUAGE,
                    "LANGUAGE_PATTERN_FAILED",
                    d,
                ),
                extra={"package_version": __version__},
            )

        # Call the modules sighup_handler which will write the configuration
        # data to the log
        log_config_data(config_parser, logger)

    def __handle_lockfile__(self):
        """[summary]

        :raises LockedInitialisationOfSingleton: [description]
        :raises LockedInitialisationOfSingleton: [description]
        """
        global i_have_lock, logger, is_testing, lock_file_name

        def init_lockfile(fq_path_to_lockfile_p):
            fp = open(lock_file_name, "w")
            fp.write("%s" % os.getpid())
            fp.close()
            logger.info(
                f"system {service_name} lockfile {lock_file_name} created",
                extra={"package_version": __version__},
            )

        if not os.path.isfile(lock_file_name):
            try:
                init_lockfile(lock_file_name)
                i_have_lock = True
            except FileNotFoundError:  # pragma: no cover
                s = f"Error opening/writing lock File {lock_file_name}"
                s += f"EXIT Now version {__version__}"
                logger.exception(s)
                sys.stderr.write(s)
                sys.stderr.write(traceback.format_exc())
                sys.exit(1)
            logger.info(
                "system %s started new lock guarded instance on %s"
                % (service_name, lock_file_name),
                extra={"package_version": __version__},
            )
        else:
            with open(lock_file_name, "r") as fp:
                pid = fp.read().strip()
            st = os.stat(lock_file_name)
            age = hms_string(time.time() - st.st_mtime)
            s = f"{lock_file_name}  locked by process with pid "
            s += f"{pid}  for: {age} : version {__version__}"
            logger.info(s)
            try:
                os.kill(int(pid), 0)
                s = f"process with pid {pid} is still alive Exit now "
                logger.info(s, extra={"package_version": __version__})
                raise LockedInitialisationOfSingleton(s)
            except OSError:
                s = f"process with pid {pid}s is not alive:\
                     Will Remove the lock file"
                logger.error(
                    s,
                    extra={"package_version": __version__},
                )
                sys.stderr.write(s)
                os.remove(lock_file_name)
                raise LockedInitialisationOfSingleton(s)

    def verbose(self):
        """[summary]

        Logging messages additionally will be sent to stderr.
        """
        global logger
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formater = logging.Formatter("%(asctime)s   %(lineno)d - %(message)s")
        ch.setFormatter(formater)
        logger.addHandler(ch)

    def __exit__(self, exc_type, exc_value, traceback):
        """[summary]

        :param exc_type: [description]
        :type exc_type: [type]
        :param exc_value: [description]
        :type exc_value: [type]
        :param traceback: [description]
        :type traceback: [type]
        """
        if exc_type == 1 and exc_value == 2 and traceback == 3:
            try:
                if os.path.isfile(lock_file_name):
                    logger.debug(
                        f"remove lock_file {lock_file_name}",
                        extra={"package_version": __version__},
                    )
                    os.unlink(lock_file_name)
            except FileNotFoundError:  # pragma: no cover
                s = f"ERR: remove lock_file: version {__version__}"
                logger.exception(s)
                sys.stdout.write(s)
                sys.stderr.write(s)
                sys.stderr.write(traceback.print_exc())
        else:  # pragma: no cover
            logger.error(
                "remove log_file Leave Singleton with nonstandard __exit__",
                extra={"package_version": __version__},
            )
            sys.stderr.write("Leave Singleton with nonstandard __exit__")

    def __del__(self):
        """Allow to use in with."""
        #        #pass
        #        # rudiments from python2 - seems to be not needed any longer
        if i_have_lock:
            self.__exit__(1, 2, 3)


class ContextFilter(logging.Filter):
    """[summary]

    Merges additional values into the logging message.

    :param logging: [description]
    :type logging: [type]
    :return: [description]
    :rtype: [type]
    """

    def filter(self, record):
        """[summary]

        Add additional field to a log message.

        :param record: [description]
        :type record: [type]
        :return: [description]
        :rtype: [type]
        """
        if "package_version" not in record.__dict__:
            record.package_version = "undef"
        record.SYSTEM_ID = os.getenv("SYSTEM_ID", "None")
        record.SUB_SYSTEM_ID = os.getenv("SUB_SYSTEM_ID", "None")
        record.SUB_SUB_SYSTEM_ID = os.getenv("SUB_SUB_SYSTEM_ID", "None")
        record.PRODUKT_ID = os.getenv("PRODUKT_ID", "None")
        record.USER_SPEC_1 = os.getenv("USER_SPEC_1", "None")
        record.USER_SPEC_2 = os.getenv("USER_SPEC_2", "None")
        return True


class LockedInitialisationOfSingleton(Exception):
    """Exception raised if the start of a service finds an existing lockfile.

    :param Exception: [description]
    :type Exception: [type]
    """

    def __init__(self, message: str):
        """Write a message.

        :param message: [description]
        :type message: [str]
        """
        self.message = message


class ForbiddenInitialisationOfSingleton(Exception):
    """Exception raised if the start of a Service is not possible.

    :param Exception: [description]
    :type Exception: [type]
    """

    def __init__(self, message):
        """[summary]

        :param message: [description]
        :type message: [type]
        """
        self.message = message


def template_writer(
    patterns_p: dict, pattern_offset_p: dict, pattern_p: dict, d_vars_p: dict
):
    """

    Template writer used to generate strings based on the Pattern module.

    :param patterns_p: [description]
    :type patterns_p: dict
    :param pattern_offset_p: [description]
    :type pattern_offset_p: dict
    :param pattern_p: [description]
    :type pattern_p: dict
    :param d_vars_p: [description]
    :type d_vars_p: dict
    :raises KeyError: [description]
    :return: [description]
    :rtype: [type]
    """
    try:
        template = patterns_p[pattern_offset_p][pattern_p]
        return template % (d_vars_p)
    except KeyError:
        logger.exception(
            "unable to generate string in %s from template %s"
            % (pattern_offset_p, pattern_p)
        )
        raise KeyError(pattern_offset_p)


def ps_shell(cmd_p: str, env_p: dict = None):
    """
    Execute a command in a spawned Process.

    :param cmd_p: [cmd to be executed ]
    :type cmd_p: str
    :param env_p: [environment to use], defaults to None
    :type env_p: dict, optional
    :return: [(out, err, exit, time_needed)]
    :rtype: [tuple]
    """
    global logger, service_name
    success = "SUCCESS"
    start_time = time.time()
    child = Popen(
        cmd_p,
        shell=True,
        stdout=PIPE,
        stderr=PIPE,
        env=env_p,
        universal_newlines=True,
    )
    (stdout, stderr) = child.communicate()
    exitcode = child.poll()
    end_time = time.time()

    time_needed = hms_string(end_time - start_time)
    if exitcode != 0:
        success = "ERROR"
    logger.debug(
        f"{success} {time_needed}  {cmd_p}",
        extra={"package_version": __version__},
    )
    if exitcode != 0:
        logger.error(
            "%d %s" % (int(exitcode), stderr),
            extra={"package_version": __version__},
        )

    return (
        str(stdout).strip().split("\n"),
        str(stderr).strip().split("\n"),
        exitcode,
        time_needed,
    )


def exec_interpreter_from_string(source_code: str):
    """Execute  a given python programm in a new spawned python interpreter.

    :param source_code: [python source code]
    :type source_code: str
    :return: [(out, err, exit, time_needed)]
    :rtype: [tuple]


    """
    tmp = tempfile.NamedTemporaryFile(mode="w+t")
    try:
        tmp.write(source_code)
        tmp.flush()
        cmd = f"{sys.executable}  {tmp.name}"
        out, err, exit, time_needed = ps_shell(cmd)
    finally:
        tmp.close()

    return out, err, exit, time_needed



def get_html_string(var_p: dict):
    """[summary]

    Return html table structure for a dictionaries structure. Makes it easy
    to display dictionaries in log output messages.

    :param var_p: [description]
    :type var_p: dict
    :return: [str]
    :rtype: [type]

    >>> get_html_string({"A":"B"})
    '<table><tr><td>A</td><td>B</td></tr></table>'
    """
    html_string = "<table>"
    for name, value in var_p.items():
        if type(value) is dict:
            html_string += "<tr><td>%s</td><td>%s</td></tr>" % (
                str(name),
                get_html_string(value),
            )
        else:
            html_string += "<tr><td>%s</td><td>%s</td></tr>" % (
                str(name),
                str(value),
            )

    html_string += "</table>"
    return html_string


def hms_string(sec_elapsed: int):
    """[summary]

    Return the hour:minute:seconds for a period in seconds.

    :param sec_elapsed: [description]
    :type sec_elapsed: int
    :return: [description]
    :rtype: [type]

    >>> hms_string(600)
    '00:10:00'

    """
    h = int(sec_elapsed / (60 * 60))
    m = int(sec_elapsed % (60 * 60) / 60)
    s = sec_elapsed % 60
    return "%02d:%02d:%02d" % (h, m, s)



"""Tests for the Config Module."""
import os
import signal
import sys
from configparser import ParsingError

from ps.basic import Config, DEV_STAGES, Patterns

import pytest

from t_utils import (
    TEST_SERVICE_NAME,
    common_Config_class_attributes_after_initialisation,
    get_data_of_file,
    reset_singleton,
)


os.environ["IS_TESTING"] = "YES"

# The fixtures:
#   - dev_allowed_stages
#   - dev_wrong_stages
# are defined in conftest.py


def test_that_an_reinitialization_of_the_singleton_raises_an_error(
    dev_allowed_stages
):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        b = Config.Basic(TEST_SERVICE_NAME)  # Should be  OK
        assert b
        Config.Basic(TEST_SERVICE_NAME)  # Will Raise Error


def test_service_names_with_wrong_stage_name_raise_error(dev_wrong_stages):
    """[summary]

    :param dev_wrong_stages: [description]
    :type dev_wrong_stages: [type]
    """
    reset_singleton()
    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        assert Config.service_name == "not_yet_defined"
        Config.Basic(TEST_SERVICE_NAME)


def test_properties_are_set(dev_allowed_stages):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    Config.Basic(TEST_SERVICE_NAME)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert Config.primary_herald_url == ""


#
# verbose mode
#
def test_verbose_mode_with_uninitialized_singleton_raises_error(
    dev_allowed_stages,
):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    with pytest.raises(TypeError):
        assert Config.service_name == "not_yet_defined"
        Config.Basic.verbose()


def test_verbose_mode_with_writes_logging_messages_to_stdout_too(
    dev_allowed_stages, capsys
):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    singleton = Config.Basic(TEST_SERVICE_NAME)
    singleton.verbose()
    msg = "This is a message"
    Config.logger.info(msg, extra={"package_version": "__version__"})
    captured = capsys.readouterr()
    assert captured.out == ""
    assert msg in captured.err
    assert msg in get_data_of_file(Config.log_file_name)


#
# CONFIG FILE THINGS
#
def test_not_existing_config_file_directory_raises_error(dev_allowed_stages,
                                                         capsys
                                                         ):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    config_file_directory = "/foobar"
    os.environ["BASIC_CONFIGFILE_DIR"] = config_file_directory
    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        Config.Basic(TEST_SERVICE_NAME, have_config_file=True)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert Config.config_file_name == "not_yet_defined"
    assert "BASIC_CONFIGFILE_DIR /foobar not found." in get_data_of_file(
        Config.log_file_name
    )


def test_config_file_in_local_directory(dev_allowed_stages, capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    os.environ["BASIC_CONFIGFILE_DIR"] = ""
    del os.environ["BASIC_CONFIGFILE_DIR"]

    config_file_name = os.path.join(
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg"
    )
    with open(config_file_name, "w") as fp:
        fp.write("[test_section]\n")
        fp.write("    test_name = test_value  \n")

    Config.Basic(TEST_SERVICE_NAME, have_config_file=True)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert Config.primary_herald_url == ""

    assert Config.config_file_name == os.path.join(os.getcwd(),
                                                   config_file_name)
    assert "test_section" in Config.config_parser.sections()
    assert "test_value" == Config.config_parser["test_section"]["test_name"]


def test_config_file_in_local_directory_not_given(dev_allowed_stages, capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    os.environ["BASIC_CONFIGFILE_DIR"] = ""
    del os.environ["BASIC_CONFIGFILE_DIR"]

    config_file_name = os.path.join(
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg"
    )
    if os.path.isfile(config_file_name):
        os.remove(config_file_name)

    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        Config.Basic(TEST_SERVICE_NAME, have_config_file=True)


def test_config_file_in_directory(dev_allowed_stages, capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    config_file_directory = "/tmp"
    config_file_name = os.path.join(
        config_file_directory,
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg",
    )
    os.environ["BASIC_CONFIGFILE_DIR"] = config_file_directory
    with open(config_file_name, "w") as fp:
        fp.write("[test_section]\n")
        fp.write("    test_name = test_value  \n")

    Config.Basic(TEST_SERVICE_NAME, have_config_file=True)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert Config.primary_herald_url == ""

    assert Config.config_file_name == config_file_name
    assert "test_section" in Config.config_parser.sections()
    assert "test_value" == Config.config_parser["test_section"]["test_name"]


def test_wrong_formated_config_files_in_directory_raise_error(
    dev_allowed_stages, capsys
):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    config_file_directory = "/tmp"
    config_file_name = os.path.join(
        config_file_directory,
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg",
    )
    os.environ["BASIC_CONFIGFILE_DIR"] = config_file_directory
    with open(config_file_name, "w") as fp:
        fp.write("[Wrong Section Entry\n")
        fp.write("    test_name = test_value  \n")

    with pytest.raises(ParsingError):
        Config.Basic(TEST_SERVICE_NAME, have_config_file=True)
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert Config.primary_herald_url == ""

    assert Config.config_file_name == config_file_name
    assert "MissingSectionHeaderError" in get_data_of_file(
        Config.log_file_name)


def test_forced_config_file_raises_error_on_impossible_pattern_language(
    dev_allowed_stages, capsys
):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    config_file_directory = "/tmp"
    config_file_name = os.path.join(
        config_file_directory,
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg",
    )
    os.environ["BASIC_CONFIGFILE_DIR"] = config_file_directory
    with open(config_file_name, "w") as fp:
        fp.write("[GLOBAL]\n")
        fp.write("    pattern_language = IMPOSSIBLE\n")

    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        Config.Basic(TEST_SERVICE_NAME, have_config_file=True)

    common_Config_class_attributes_after_initialisation(dev_allowed_stages)


def test_forced_config_file_reads_herald_url_and_pattern_language(
    dev_allowed_stages, capsys
):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    config_file_directory = "/tmp"
    config_file_name = os.path.join(
        config_file_directory,
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg",
    )
    os.environ["BASIC_CONFIGFILE_DIR"] = config_file_directory
    with open(config_file_name, "w") as fp:
        fp.write("[GLOBAL]\n")
        fp.write("    herald_url = localhost\n")
        fp.write("    pattern_language = DE\n")

    Config.Basic(
        TEST_SERVICE_NAME,
        have_config_file=True,
        have_herald_url_in_config_file=True,
    )
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)
    assert Config.primary_herald_url == "localhost"

    assert Config.config_file_name == config_file_name
    assert "GLOBAL" in Config.config_parser.sections()
    assert "localhost" == Config.config_parser["GLOBAL"]["herald_url"]
    assert "DE" == Config.config_parser["GLOBAL"]["pattern_language"]
    assert "GLOBAL" in get_data_of_file(Config.log_file_name)
    assert "herald_url" in get_data_of_file(Config.log_file_name)
    assert "localhost" in get_data_of_file(Config.log_file_name)


def test_forced_config_file_but_herald_url_is_not_defined(dev_allowed_stages,
                                                          capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    config_file_directory = "/tmp"
    config_file_name = os.path.join(
        config_file_directory,
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg",
    )
    os.environ["BASIC_CONFIGFILE_DIR"] = config_file_directory
    with open(config_file_name, "w") as fp:
        fp.write("[GLOBAL]\n")
        fp.write("    pattern_language = DE\n")
        fp.write("    LOGGING =  INFO\n")

    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        Config.Basic(
            TEST_SERVICE_NAME,
            have_config_file=True,
            have_herald_url_in_config_file=True,
        )
    common_Config_class_attributes_after_initialisation(dev_allowed_stages)

    assert "Value of herald_url" in get_data_of_file(Config.log_file_name)
    assert "not given" in get_data_of_file(Config.log_file_name)


#
# SIGHUP HANDLER THINGS
#
def test_sighup_handler_singleton_uninitialized(dev_allowed_stages, capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    with pytest.raises(Config.ForbiddenInitialisationOfSingleton):
        os.kill(os.getpid(), signal.SIGHUP)


def test_sighup_handler_singleton_initialized(dev_allowed_stages, capsys):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    :param capsys: [description]
    :type capsys: [type]
    """
    reset_singleton()
    os.environ["BASIC_CONFIGFILE_DIR"] = ""
    del os.environ["BASIC_CONFIGFILE_DIR"]

    config_file_name = os.path.join(
        TEST_SERVICE_NAME + DEV_STAGES[dev_allowed_stages]["suffix"] + ".cfg"
    )
    with open(config_file_name, "w") as fp:
        fp.write("[test_section]\n")
        fp.write("    test_name = test_value  \n")
        fp.write("[GLOBAL]\n")
        fp.write("    LOGGING =  DEBUG\n")

    Config.Basic(TEST_SERVICE_NAME, have_config_file=True)
    os.kill(os.getpid(), signal.SIGHUP)
    assert "SIGHUP Received. Print Configuration." in get_data_of_file(
        Config.log_file_name
    )
    print(get_data_of_file(Config.log_file_name))
    print(get_data_of_file(Config.config_file_name))


#
# ps_shell THINGS
#
def test_ps_shell_ls(dev_allowed_stages):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    Config.Basic(TEST_SERVICE_NAME)
    out, err, exit_code, time_needed = Config.ps_shell("ls -a")
    assert "LOG" in out
    assert err == [""]
    assert exit_code == 0
    assert "ls -a" in get_data_of_file(Config.log_file_name)


def test_ps_shell_impossible_cmd(dev_allowed_stages):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    Config.Basic(TEST_SERVICE_NAME)
    out, err, exit_code, time_needed = Config.ps_shell("impossible_command")
    assert out == [""]
    assert "impossible_command" in "".join(err)
    assert exit_code != 0
    assert "impossible_command" in get_data_of_file(Config.log_file_name)


#
# exec_interpreter THINGS
#
def test_exec_interpreter_from_string(dev_allowed_stages):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    Config.Basic(TEST_SERVICE_NAME)
    out, err, exit_code, time_needed = Config.exec_interpreter_from_string(
        "print('Hello World')"
    )
    assert "Hello World" in out
    assert err == [""]
    assert exit_code == 0
    assert sys.executable in get_data_of_file(Config.log_file_name)


#
# template_writer THINGS
#
def test_template_writer(dev_allowed_stages):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    Config.Basic(TEST_SERVICE_NAME)
    text = Config.template_writer(
        Patterns.SUBJECT_PATTERNS,
        "DE",
        "PROD_LOCKED",
        {"snapshot_name": "tralala"},
    )
    assert text == "tralala LOCKED "


def test_template_writer_(dev_allowed_stages):
    """[summary]

    :param dev_allowed_stages: [description]
    :type dev_allowed_stages: [type]
    """
    reset_singleton()
    Config.Basic(TEST_SERVICE_NAME)

    with pytest.raises(KeyError):
        Config.template_writer(
            Patterns.SUBJECT_PATTERNS,
            "IMPOSSIBLE_KEY",
            "PROD_LOCKED",
            {"snapshot_name": "tralala"},
        )
    assert "unable to generate strin" in get_data_of_file(Config.log_file_name)



"""ps.basic package."""
import logging

import pkg_resources

from .utils import get_html_string, hms_string  # noqa: F401

# get the version number defined in setup.py
__version__ = pkg_resources.get_distribution("ps.basic").version


DEV_STAGES = {
    "TESTING": {
        "suffix": "_t",
        "logging_port": 9010,
        "logging_bridge_port": 9011,
        "logging_level": logging.DEBUG,
        "l_admin_mail": ["test@somewhere.com"],
    },
    "DEVELOPMENT": {
        "suffix": "_d",
        "logging_port": 9020,
        "logging_bridge_port": 9019,
        "logging_level": logging.DEBUG,
        "l_admin_mail": ["d_test@somewhere.com"],
    },
    "INTEGRATION": {
        "suffix": "_i",
        "logging_port": 9022,
        "logging_bridge_port": 9018,
        "logging_level": logging.DEBUG,
        "l_admin_mail": ["itest@somewhere.com"],
    },
    "PRODUCTION": {
        "suffix": "",
        "logging_port": 9024,
        "logging_bridge_port": 9017,
        "logging_level": logging.DEBUG,
        "l_admin_mail": ["production@somewhere.com"],
    },
}



"""
Implement the ps Finite State Machine.

 >>> import os
"""


import ps.basic.Config
from ps.basic import __version__ as version
from ps.basic import get_html_string
from ps.basic.State import State


class TransitionError(Exception):
    """[summary]

    State Exception.

    :param Exception: [description]
    :type Exception: [type]
    """

    def __init__(self, message):
        """[summary]

        :param message: [description]
        :type message: [type]
        """
        self.message = message


class FiniteStateMachine(object):
    """[summary]

         :param object: [description]
         :type object: [type]

         Import ...

         >>> from ps.basic.State import State, StateError
         >>> from ps.basic.fsm import FiniteStateMachine, TransitionError

         Define  handler functions.

         The __doc__ variables text of a handler function lateron will
         be displayed within the generated documentations tooltip.

         >>> def init_state_handler(state: State, context_p: dict) -> str:
         ...   __doc__="This is the documentation of init_state_handler"
         ...   context_p[state.name] = "visited"
         ...   if True:
         ...       return "INIT_OK"
         ...   else:
         ...       return "DIRECT"
         >>> def state1_handler(state: State, context_p: dict) -> str:
         ...   __doc__="This is the documentation of state1_state_handler"
         ...   context_p[state.name] = "visited"
         ...   return "STATE1_OK"
         >>> def error_state_handler(state: State, context_p: dict) -> str:
         ...   __doc__="This is the documentation of error_state_handler"
         ...   context_p[state.name] = "visited"
         ...   return "ERROR_STATE_OK"
         >>> def fin_state_handler(state: State, context_p: dict) -> str:
         ...   __doc__="This is the documentation of fin_state_handler"
         ...   context_p[state.name] = "visited"
         ...   return "FIN_STATE_OK"

        Given the defined handler_functions, we are able to define State
        variables using those handler functions.

        First a final state is defined - the fsm terminates if it enters
        a final state.

         >>> final = State("FINAL", fin_state_handler, final=True)

        Next we define an error state - entering the error state may result
        in sending an email to some recipients - defined in the mail_addr list.

         >>> error = State( "ERROR",error_state_handler, error=True,
         ...     mail_addr='[{"A":"ror_handler@mail.company"}]',
         ...     default=final,)

        Now we define state handling functions - the default parameter
        tells the state, that in case  the states compute_function
        returns a string not definied
        im the state's transition strings, that state will be entered.

         >>> start = State("START", init_state_handler,
         ...                initial=True, default=error)
         >>> state1 = State("STATE1", state1_handler, default=error)


         Currently we have 4 State Instances named final, error,
         start and state1.

         A State instance  (inherited from dict)  defines a
         key/value pair where:

             - key is a string (the string returned by a
               states computing function)
             - value is the reference to the state to be
               visited  if the states
               computing function returns that string.

         - if in state start  and start.computing_function alia
           init_state_handler() returns "INIT_OK" we switch
           to state state1

         >>> start["INIT_OK"] = state1
         >>> start["DIRECT"] = final

             - if in state state1 and  state1.computing_function()
               alia state1_state_handler() returns "STATE1_OK"
               we switch to state final

         >>> state1["STATE1_OK"] = final

             - if in state error  and error.computing_function()
               alia error_state_handler() returns
               "ERROR_STATE_OK" we switch to state final

         >>> error["ERROR_STATE_OK"] = final

         Creating an instance of the Finite State machine

         >>> fsm = FiniteStateMachine("ThedoctestMachine")

         Add the before defined states and state exhchanges
         to the FiniteStateMachine instance *fsm*

         >>> fsm.add_state([start, state1, error, final])

         Define an empty context

         >>> context = {}

         Run the finite State machine.

         >>> from ps.basic import Config
         >>> import os
         >>> os.environ["DEV_STAGE"] = "TESTING"
         >>> singleton=Config.Basic("fsm_test")
         >>> fsm.run(context)

         As the (for all states) used handler_function  puts its
         state name to the context we get an context as assumed.

         >>> context
         {'START': 'visited', 'STATE1': 'visited', 'FINAL': 'visited'}

        To Check the behaviour of the fsm we can inspect the trace of
        the last run - it is remebered in fsm.inputs:

        >>> fsm.inputs
        ['INIT_OK', 'STATE1_OK', 'FIN_STATE_OK']

        For the next run we are able to  Reset the fsms state, the
        inputs will be empty.

        >>> fsm.reset()
        >>> fsm.inputs
        []

        To check  the transitions defined  inside the fsm:

        >>> fsm.all_transitions # doctest:+ELLIPSIS
        [(<'START' State @ 0x...>, 'INIT_OK', <'STATE1' State @ 0x...>)]

    ..    [(<'START' State @ 0x...>, 'INIT_OK', <'STATE1' State @ 0x...>),
          (<'STATE1' State @ 0x...>, 'STATE1_OK', <'FINAL' State @ 0x...>),
          (<'ERROR' State @ 0x...>, 'ERROR_STATE_OK', <'FINAL' State @ 0x...>)]


        Generate an image of the fsm - based on it's implementation ...

        >>> from ps.basic.get_graph import get_graph
        >>> graph = get_graph(fsm)
        >>> graph.draw(f"{fsm.name}.svg", prog="dot")

    .. Comment  Copy the file to the documentation environment
       >>> assert os.path.isfile(f"{fsm.name}.svg")
       >>> from shutil import copyfile
       >>> copyfile(f"{fsm.name}.svg", f"../docs/source/{fsm.name}.svg")
       '../docs/source/ThedoctestMachine.svg'

    It serves as a good documentation for the implementatons behaviour.

        .. image:: ThedoctestMachine.svg

    """

    def __init__(self, name: str):
        """[summary]

        Construct a FSM.

        :param name: [description]
        :type name: str
        """
        self.name = name
        self.current_state, self.init_state = None, None
        self.inputs, self.states, = (
            [],
            [],
        )
        self.final_states = []
        self.error_states = []

    def add_state(self, states: [State]) -> None:
        """[summary]

        Add State(s) to the fsm.

        :param states: [description]
        :type states: [type]
        :raises TransitionError: [description]
        """
        for state in states:
            if state.error:
                self.error_states.append(state)
            if state.final:
                self.final_states.append(state)
            if state.initial:
                if self.init_state is None:
                    self.init_state = state
                else:
                    raise TransitionError("Only One initial State allowed.")
            self.states.append(state)

    @property
    def all_transitions(self):
        """[summary]

        Get transitions from the states of the fsm.

        :return: [ (source state, input, destination state) ]
        :rtype: [type]
        """
        transitions = list()
        for src_state in self.states:
            for input_value, dst_state in src_state.items():
                transitions.append((src_state, input_value, dst_state))
        return transitions

    def transition(self, input_value: str):
        """[summary]

        Transition to the next state.

        :param input_value: [description]
        :type input_value: str
        """
        current = self.current_state
        destination_state = current[input_value]
        self.current_state = destination_state

    def reset(self):
        """[summary]

        Enter the Finite State Machine.
        """
        self.current_state = self.init_state
        del self.inputs
        self.inputs = list()

    def process(self, input_data: str):
        """[summary]

        Process input data.

        :param input_data: [description]
        :type input_data: str
        """
        self.reset()
        for item in input_data:
            self.transition(item)

    def run(self, context: dict):
        """[summary]

        Process input data.

        :param context: [description]
        :type context: dict
        """
        self.reset()
        current_state = self.init_state
        ps.basic.Config.logger.debug(
            "FSM: " + self.name + " started in State " + str(current_state),
            extra={"package_version": version},
        )
        while True:
            my_current_state = self.current_state
            rval = my_current_state.compute_function(my_current_state, context)
            s = f"{my_current_state.name}-compute_function ret {str(rval)}"
            # print(s)
            ps.basic.Config.logger.debug(
                s,
                extra={"package_version": version},
            )
            s = f"{self.name}.run new step: {my_current_state.name} compute"
            s += f" function left context as {get_html_string(context)}"
            ps.basic.Config.logger.debug(
                s,
                extra={"package_version": version},
            )
            self.inputs.append(rval)
            if my_current_state.final is True:
                ps.basic.Config.logger.debug(
                    f"""Leave run, as
                     {self.current_state.name}
                      is a final state """,
                    extra={"package_version": version},
                )
                break
            else:
                self.transition(rval)
                ps.basic.Config.logger.debug(
                    f"""TRANSITION: from State
                    {my_current_state.name}
                    <br> ONInput:
                    {str(rval)}
                    <br>To State:
                    {self.current_state.name}""",
                    extra={"package_version": version},
                )



"""Generic linux daemon base class for python 3.x."""

import atexit
import os
import signal
import sys
import time


class Daemon:
    """A generic daemon class.

    import os
    import sys
    from ps.basic.daemon import daemon
    from ps.herald.ps_bridge import main
    os.environ["DEV_STAGE"] = "DEVELOPMENT"

    class MyDaemon(daemon):
        def run(self):
            sys.argv[0] = "ps_bridge"
            sys.argv[1] = "-v"
            sys.stdout = open('ps_bridge.olog', 'a+')
            sys.stderr = open('ps_bridge.elog', 'a+')
            main()


    if __name__ == "__main__":
        daemon = MyDaemon("/tmp/ps_bridge.pid")
        if len(sys.argv) == 2:
            if "start" == sys.argv[1]:
                daemon.start()
            elif "stop" == sys.argv[1]:
                daemon.stop()
            elif "restart" == sys.argv[1]:
                daemon.restart()
            else:
                print("Unknown command")
                sys.exit(2)
            sys.exit(0)
        else:
            print("usage: %s start|stop|restart" % sys.argv[0])
            sys.exit(2)
    """

    def __init__(self, pidfile):
        """Set the name of the pidfile."""
        self.pidfile = pidfile

    def daemonize(self):
        """Deamonize the process."""
        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as err:
            sys.stderr.write(f"fork #1 failed: {err}\n")
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write(f"fork #2 failed: {err}\n")
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        fd_in = open(os.devnull, "r")
        fd_out = open(os.devnull, "a+")
        fd_err = open(os.devnull, "a+")

        os.dup2(fd_in.fileno(), sys.stdin.fileno())
        os.dup2(fd_out.fileno(), sys.stdout.fileno())
        os.dup2(fd_err.fileno(), sys.stderr.fileno())

        # create pidfile
        pid = str(os.getpid())
        with open(self.pidfile, "w+") as fp:
            fp.write(pid + "\n")

        # register del_pid to be called on exit
        atexit.register(self.delpid)

    def delpid(self):
        """Delete the pidfile."""
        os.remove(self.pidfile)

    def start(self):
        """Start the daemon."""
        # Check for a pidfile to see if the daemon already runs
        try:
            with open(self.pidfile, "r") as fp:
                pid = int(fp.read().strip())
        except IOError:
            pid = None

        if pid:
            message = f"pidfile {self.pidfile} already exist. "
            message += "Daemon already running?\n"
            sys.stderr.write(message)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """Stop the daemon."""
        # Get the pid from the pidfile
        try:
            with open(self.pidfile, "r") as fp:
                pid = int(fp.read().strip())
        except IOError:
            pid = None

        if not pid:
            message = f"pidfile {self.pidfile} does not exist. "
            message += "Daemon not running?\n"
            sys.stderr.write(message)
            return

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            e = str(err.args)
            if e.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err.args))
                sys.exit(1)

    def restart(self):
        """Restart the daemon."""
        self.stop()
        self.start()

    def run(self):
        """Xecute the provided function.

        You should override this method when you subclass Daemon.
        It will be called after the process has been daemonized by
        start() or restart().
        """



"""Setup."""

from setuptools import setup


version = "1.2.11"

INSTALL_REQUIRES = ["setuptools", "docopt", "pygraphviz"]

setup(
    name="ps.basic",
    version=version,
    description="Basic class used in PS environment",
    long_description=open("README.rst").read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: Other/Proprietary License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Utilities",
    ],
    keywords="PS",
    author="Setz",
    author_email="thomas@setz.de",
    url="https://bitbucket.org/drsetz/ps.basic",
    license="GPL",
    packages=["ps.basic"],
    package_dir={"": "src"},
    # include_package_data=True,
    zip_safe=False,
    install_requires=INSTALL_REQUIRES,
    entry_points={
        "console_scripts": [
            "ps_shell_log=ps.basic.ps_shell_log:main",
        ],
    },
)
